########################################################################
# Copyright (c) 2018
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# Steffen Enders <steffen<at>enders<dot>nrw>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import struct
import os
import json
import logging
import re
import math
from operator import itemgetter
from itertools import groupby

# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOG = logging.getLogger(__name__)

NUMPY_AVAILABLE=False
try:
    import numpy as np
    NUMPY_AVAILABLE = True
except:
    LOG.warning("numpy/scipy not available, which could otherwise speed up matching.")


class ApiVector(object):

    def __init__(self, winapi1024_filepath=None, sort_vector=True):
        self._sort_vector = sort_vector
        self._winapi1024 = self._loadWinApi1024(winapi1024_filepath)
        self._dllapi_only = list(zip(map(itemgetter(0), self._winapi1024), map(itemgetter(1), self._winapi1024)))
        # linear
        self._vector_ranks_only = [entry[3] for entry in self._winapi1024]
        self._n_vector_ranks_only = []
        if NUMPY_AVAILABLE:
            self._n_vector_ranks_only = np.array(self._vector_ranks_only)
        # equal
        # self._vector_ranks_only = [1 for entry in self._winapi1024]
        # sigmoid
        # self._vector_ranks_only = [int(100 * round((1 + math.tanh(3.0 * (rank - 512) / 1024)) / 2, 2)) for rank in self._vector_ranks_only]
        self._base64chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz@}]^+-*/?,._"
        self._bin2base64 = {"{:06b}".format(i): base64char for i, base64char in enumerate(self._base64chars)}
        self._base642bin = {v: k for k, v in self._bin2base64.items()}

    def _loadWinApi1024(self, winapi1024_filepath):
        winapi1024 = []
        if winapi1024_filepath:
            if os.path.isfile(winapi1024_filepath):
                with open(winapi1024_filepath, "r") as infile:
                    for line in sorted(infile.readlines(), key=lambda line: int(line.split(";")[3].strip()), reverse=True) if self._sort_vector else infile.readlines():
                        functionality = line.split(";")[0]
                        dll, function = line.split(";")[2].strip().split("!")
                        rank = int(line.split(";")[3].strip())
                        winapi1024.append((dll, function, functionality, rank))
                if len(winapi1024) != 1024:
                    LOG.warn("WinApi1024 file contained {} instead of 1024 api definitions.".format(len(winapi1024)))
            else:
                LOG.error("Not a file: %s!", winapi1024_filepath)
                raise ValueError
        return winapi1024
        
    def _loadCollectionData(self, collection_filepath):
        vectors_by_family = {}
        if collection_filepath:
            if os.path.isfile(collection_filepath):
                with open(collection_filepath, "r") as infile:
                    for line in infile.readlines():
                        family = line.split(";")[0]
                        sample_path = line.split(";")[1].strip()
                        compressed_api_vector = line.split(";")[4].strip()
                        if not family in vectors_by_family:
                            vectors_by_family[family] = {}
                        vectors_by_family[family][sample_path] = compressed_api_vector
            else:
                LOG.error("Not a file: %s!", collection_filepath)
                raise ValueError
        return vectors_by_family
        
    def getWinApi1024(self):
        return self._winapi1024
   
    def getApiVectors(self, results):
        unique_results = self._get_uniquified_results(results)
        api_vectors = self._vectorize(results)
        coverage = {}
        for api_map_name in unique_results:
            percentage = 100.0 * sum(api_vectors[api_map_name]) / len(unique_results[api_map_name]) if len(unique_results[api_map_name]) else 0
            coverage[api_map_name] = {
                "vector": self.compress(api_vectors[api_map_name]),
                "num_unique_apis": len(unique_results[api_map_name]), 
                "in_api_vector": sum(api_vectors[api_map_name]), 
                "percentage": percentage
            }
        return coverage

    def getApiVectorFromApiList(self, api_list):
        scout_format = {
            "user_list": []
        }
        for index, list_entry in enumerate(api_list):
            scout_list_entry = [
                0,  # unknown offset
                0,  # unknown api_address
                list_entry.split("!")[0] + "_stubbedBitness", 
                list_entry.split("!")[1], 
                index + 1,  # IAT offset
                1  # ref count
            ]
            scout_format["user_list"].append(scout_list_entry)
        return self.getApiVectors(scout_format)
        
    def getListFromApiVector(self, vector):
        if not isinstance(vector, list):
            vector = self.decompress(vector)
        api_list = []
        for index, entry in enumerate(self._winapi1024):
            if vector[index]:
                api_list.append("{}!{}".format(entry[0], entry[1]))
        return sorted(api_list)
        
    def getApiVectorFromApiDictionary(self, api_dict):
        scout_format = {
            "user_list": []
        }
        index = 0
        for dll_entry in api_dict:
            for api_entry in api_dict[dll_entry]:
                scout_list_entry = [
                    0,  # unknown offset
                    0,  # unknown api_address
                    dll_entry + "_stubbedBitness", 
                    api_entry, 
                    index + 1,  # IAT offset
                    1  # ref count
                ]
                scout_format["user_list"].append(scout_list_entry)
                index += 1
        return self.getApiVectors(scout_format)
        
    def getDictionaryFromApiVector(self, vector):
        if not isinstance(vector, list):
            vector = self.decompress(vector)
        api_dict = {}
        for index, entry in enumerate(self._winapi1024):
            if vector[index]:
                if not entry[0] in api_dict:
                    api_dict[entry[0]] = []
                api_dict[entry[0]].append(entry[1])
        return api_dict
        
    def getVectorConfidence(self, vector):
        if len(vector) != 1024:
            vector = self.decompress(vector)
        scores = []
        for index, entry in enumerate(self._winapi1024):
            if vector[index]:
                scores.append(entry[3])
        score = 0.0
        if sum(vector):
            # confidence is calculated based on APIs less common than top75 and total number of APIs in the vector
            score = 100.0 * math.sqrt(1.0 * sum([1 for value in scores if value > 64]) / sum(vector)) * (1.0 * min(sum(vector), 20) / 20)
        return score

    def compress(self, api_vector):
        bin_vector_string = "".join(["%d" % bit for bit in api_vector]) + "00" * (len(api_vector) % 3)
        uncompressed_b64 = "".join(self._bin2base64[chunk] for chunk in self._chunks(bin_vector_string, 6))
        compressed_b64 = "".join(self._compress_rep(c, r) for c, r in groupby(uncompressed_b64))
        return compressed_b64

    def decompress(self, compressed_vector):
        if NUMPY_AVAILABLE:
            return self.n_decompress(compressed_vector)
        decompressed_b64 = "".join(self._decompress_get(compressed_vector))
        vectorized = "".join(self._base642bin[c] for c in decompressed_b64)
        padding_length = len(vectorized) - len(self._winapi1024)
        as_binary = [int(i) for i in vectorized[:-padding_length]]
        return as_binary
    
    def n_decompress(self, compressed_vector):
        decompressed_b64 = "".join(self._decompress_get(compressed_vector))
        vectorized = "".join(self._base642bin[c] for c in decompressed_b64)
        padding_length = len(vectorized) - len(self._winapi1024)
        as_binary = np.fromiter(vectorized[:-padding_length], int)
        return as_binary
    
    def _isDecompressed(self, vector):
        if NUMPY_AVAILABLE:
            return isinstance(vector, np.ndarray)
        return isinstance(vector, list)

    def matchVectors(self, vector_a, vector_b):
        # ensure binary representation and apply weights
        if not self._isDecompressed(vector_a):
            vector_a = self.decompress(vector_a)
        vector_a = self._apply_weights(vector_a)
        if not self._isDecompressed(vector_b):
            vector_b = self.decompress(vector_b)
        vector_b = self._apply_weights(vector_b)
        # calculate Jaccard index
        if NUMPY_AVAILABLE:
            maxPQ = np.sum(np.maximum(vector_a, vector_b))
            return 1.0 * np.sum(np.minimum(vector_a, vector_b)) / maxPQ
        intersection_score = 0
        union_score = 0
        jaccard_index = 0
        for offset in range(len(vector_a)):
            intersection_score += vector_a[offset] & vector_b[offset]
            union_score += vector_a[offset] | vector_b[offset]
        if union_score > 0:
            jaccard_index = 1.0 * intersection_score / union_score
        return jaccard_index
        
    def matchVectorCollection(self, vector, collection_filepath):
        collection_data = self._loadCollectionData(collection_filepath)
        results = {
            "vector": vector,
            "confidence": self.getVectorConfidence(vector),
            "collection_filepath": collection_filepath,
            "families_in_collection": len(collection_data),
            "vectors_in_collection": sum([len(samples) for family, samples in collection_data.items()])
        }
        vector_collection_results = []
        decompressed_vector = vector
        if not self._isDecompressed(vector):
            decompressed_vector = self.decompress(vector)
        for family, samples in collection_data.items():
            for sample, sample_vector in samples.items():
                vector_collection_results.append((family, sample, self.matchVectors(decompressed_vector, sample_vector)))
        results["match_results"] = sorted(vector_collection_results, key=lambda tup: tup[2], reverse=True)
        return results

    def _chunks(self, l, n):
        for i in range(0, len(l), n):
            yield l[i:i + n]

    def _compress_rep(self, char, reps):
        size = len(list(reps))
        if size <= 2: return char * size
        return "{}{:d}".format(char, size)
        
    def _decompress_get(self, data):
        for match in re.finditer(r"(?P<char>.)((?P<count>\d+))?", data):
            if not match.group("count"): yield match.group("char")
            else: yield match.group("char") * int(match.group("count"))

    def _vectorize(self, results):
        unique_results = self._get_uniquified_results(results)
        vectors = {}
        for api_map_name, unique_results in unique_results.items():
            vectors[api_map_name] = [1 if entry in unique_results else 0 for entry in self._dllapi_only]
        return vectors

    def _get_uniquified_results(self, results):
        uniquified = {}
        for api_map_name, map_results in results.items():
            uniquified[api_map_name] = set(self._uniquify_api_name("_".join(t[2].split("_")[:-1]), t[3]) for t in map_results)
        return uniquified

    def _uniquify_api_name(self, dll_name, api_name):
        if dll_name.startswith("msvcr"):
            dll_name = "msvcrt.dll"
        api_name = api_name.split("@")[0]
        if api_name.endswith("A") or api_name.endswith("W"):
            api_name = api_name[:-1]
        return dll_name, api_name

    def _apply_weights(self, vector):
        if NUMPY_AVAILABLE:
            return self._n_apply_weights(vector)
        return [f1 * f2 for f1, f2 in zip(vector, self._vector_ranks_only)]

    def _n_apply_weights(self, vector):
        return np.multiply(vector, self._n_vector_ranks_only)



import struct


_UNSIGNED_UNPACK_FORMATS = {
    2: "H",
    4: "I",
    8: "Q"
}


def get_word(buf, start):
    return _get_binary_data(buf, start, 2)


def get_dword(buf, start):
    return _get_binary_data(buf, start, 4)


def get_qword(buf, start):
    return _get_binary_data(buf, start, 8)


def _get_binary_data(buf, start, length):
    if length not in _UNSIGNED_UNPACK_FORMATS:
        raise RuntimeError("Unsupported data length")
    return struct.unpack(_UNSIGNED_UNPACK_FORMATS[length], buf[start:start + length])[0]


def get_string(buf, start_offset):
    string = ""
    current_offset = 0
    buf = bytearray(buf)
    current_byte = buf[start_offset]
    while current_byte != 0 and start_offset + current_offset < len(buf) - 1:
        string += chr(current_byte)
        current_offset += 1
        current_byte = buf[start_offset + current_offset]
    return string



########################################################################
# Copyright (c) 2018
# Steffen Enders <steffen<at>enders<dot>nrw>
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import operator
import numpy as np
import math
import numbers
import os
from PIL import Image

from .ApiQRContext import ApiQRContext
from .ApiVector import ApiVector


class ApiQR:
    """
    Class for a two-dimensional vector where each component stands for one distinct API.
    The vector is saved one-dimensionally though.
    """
    vector = None
    context = None

    def __init__(self, winapi1024_filepath, vector=None):
        self._apivector = ApiVector(winapi1024_filepath, sort_vector=False)
        self._apivector_path = winapi1024_filepath
        self.context = ApiQRContext(self._apivector.getWinApi1024())
        if vector is None:
            self.vector = self.context.empty_vector
        else:
            self.vector = vector
        
    def setVector(self, vector):
        self.vector = self._apivector.decompress(vector)

    @property
    def vector_unsorted(self):
        result = [None] * len(self.vector)
        with open(self._apivector_path, "r") as infile:
            translation = {
                new_index: entry[0]
                for new_index, entry in enumerate(sorted(enumerate(infile.readlines()), key=lambda entry: int(entry[1].split(";")[3].strip()), reverse=True))
            }
        for index, entry in enumerate(self.vector):
            result[translation[index]] = entry
        return result

    def getPng(self, destination_path, scale_factor=5):
        # colored_vector = list(map(self.__mapColor, self.vector_unsorted, self.context.colors))
        colored_vector = []
        for index in range(len(self.vector_unsorted)):
            value, raw_color = self.vector_unsorted[index], self.context.colors[index]
            color = self.__mapColor(value, raw_color)
            if color == self.context.colors_white: 
                color = self.__mapColor(0.2, raw_color)
            else:
                value = 0.4 + (value * (1 - 0.4))
                color = self.__mapColor(value, raw_color)
            colored_vector.append(color)
        scaled_vector = sum(([e] * 4 ** scale_factor for e in colored_vector), [])
        transformed_vector = np.int8(self.__vectorToHilbert(scaled_vector))
        image = Image.fromarray(transformed_vector, mode=self.context.colors_format)
        return image

    def exportPng(self, destination_path, scale_factor=5):
        image = self.getPng(destination_path, scale_factor=scale_factor)
        image.save(destination_path, format="PNG", compress_level=0)

    def getHtmlHeader(self):
        hilbert_size = int(len(self.vector_unsorted) ** 0.5)
        return """\
        <meta charset="utf-8">
        <link href="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous">
        <script src="https://code.jquery.com/jquery-3.3.1.min.js" integrity="sha256-FgpCb/KJQlLNfOu91ta32o/NMZxltwRo8QtmkMRdAu8=" crossorigin="anonymous"></script>
        <script src="https://maxcdn.bootstrapcdn.com/bootstrap/4.0.0/js/bootstrap.bundle.min.js" integrity="sha384-feJI7QwhOS+hwpX2zkaeJQjeiwlhOP+SdQDqhgvvo1DsjtiSQByFdThsxO669S2D" crossorigin="anonymous"></script>
        <style>
        td.apicell {{
            position: relative;
            width: {:.4f}%;
        }}
        td.apicell-white {{
            background-color: rgb({white[0]:d}, {white[1]:d}, {white[2]:d});
        }}
        td.apicell:after {{
            content: '';
            display: block;
            margin-top: 100%;
        }}
        td.apicell .content {{
            position: absolute;
            top: 0;
            bottom: 0;
            left: 0;
            right: 0;
        }}
        </style>\n""".format(1.0 / hilbert_size, white=self.context.colors_white)

    def _getHtmlTableContent(self, hilbert_size, vector):
        result = ""
        for hilbert_row in self.__hilbertCurve(hilbert_size):
            result += "<tr>\n"
            for index in hilbert_row:
                value, raw_color = vector[index], self.context.colors[index]
                color = self.__mapColor(value, raw_color)
                if color == self.context.colors_white: 
                    # result += '<td class="api-cell api-cell-white"></td>'
                    color = self.__mapColor(0.2, raw_color)
                    text = ":".join(self.context.apis[index])
                    result += '<td data-toggle="popover" data-content="{popover:s}" data-delay="200" class="apicell" style="background-color: rgb({color[0]:d}, {color[1]:d}, {color[2]:d});"></td>'.format(popover=text, color=color)
                else:
                    text = ":".join(self.context.apis[index])
                    # rescale value to range 0.4-1.0
                    value = 0.4 + (value * (1 - 0.4))
                    color = self.__mapColor(value, raw_color)
                    result += '<td data-toggle="popover" data-content="{popover:s}" class="apicell" style="background-color: rgb({color[0]:d}, {color[1]:d}, {color[2]:d});"></td>'.format(popover=text, color=color)
                result += "\n"
            result += "</tr>\n"
        return result

    def getHtmlTable(self):
        vector = self.vector_unsorted
        hilbert_size = int(len(vector) ** 0.5)
        result = "<table>\n"
        result += self._getHtmlTableContent(hilbert_size, vector)
        result += "</table>\n"
        result += "<script>$('td.apicell').popover({trigger: 'hover'})</script>"
        return result

    def exportHtml(self, output_path, full=False):
        result = self.getHtmlTable()
        if not full: return result
        this_path = os.path.abspath(os.path.join(os.path.dirname(__file__)))
        with open(os.sep.join([this_path, "..", "data", "html_frame.html"]), "r") as f_html:
            compressed_vector = self._apivector.compress(self.vector_unsorted)
            result = f_html.read().format(vector=compressed_vector, body=result, head=self.getHtmlHeader())
        with open(output_path, "w") as f_out:
            f_out.write(result)

    def __mapColor(self, value, color):
        if value is True or value == 1 or value == 1.0: return color
        if value is False or value == 0 or value == 0.0: return self.context.colors_white
        alpha = max(min(value, 1), 0)
        return (
            int((1 - alpha) * 255 + alpha * color[0]),
            int((1 - alpha) * 255 + alpha * color[1]),
            int((1 - alpha) * 255 + alpha * color[2]),
        )

    @property
    def vector_hilbert(self):
        return self.__vectorToHilbert(self.vector_unsorted)

    def __vectorToHilbert(self, vector):
        element_size = 1
        if len(np.array(vector)[0].shape): element_size = np.array(vector)[0].shape[0]
        hilbert_indices = self.__hilbertCurve(int(len(vector) ** 0.5))
        return np.array(vector)[np.digitize(hilbert_indices.ravel(), list(range(len(vector))), right=True)]\
            .reshape(hilbert_indices.shape[0], hilbert_indices.shape[1], element_size)

    def __hilbertCurve(self, n):
        """
        Generate Hilbert curve indexing for (n, n) array. 'n' must be a power of two.
        https://github.com/znah/notebooks/blob/master/hilbert_curve.ipynb
        """
        if n == 1: return np.zeros((1, 1), np.int32)
        t = self.__hilbertCurve(n//2)
        a = np.flipud(np.rot90(t))
        b = t + t.size
        c = t + t.size*2
        d = np.flipud(np.rot90(t, -1)) + t.size*3
        return np.vstack(map(np.hstack, [[a, b], [d, c]]))  

    def __add__(self, other):
        if not isinstance(other, ApiQR):
            raise ValueError("Only another ApiQR instance can be added.")
        if self.context != other.context:
            raise ValueError("Both ApiQR instances need to share the same context in order to be added together!")
        return ApiQR(vector=list(map(operator.add, self.vector, other.vector)), context=self.context)

    def __mul__(self, other):
        if not isinstance(other, numbers.Number):
            raise ValueError("An ApiQR instance can only be multiplied with numbers.")
        return ApiQR(vector=list(map(lambda x: x * other, self.vector)), context=self.context)

    def __div__(self, other):
        if not isinstance(other, numbers.Number):
            raise ValueError("An ApiQR instance can only be divided by numbers.")
        return ApiQR(vector=list(map(lambda x: 1.0 * x / other, self.vector)), context=self.context)

    def __truediv__(self, other):
        return self.__div__(other)

    def __str__(self):
        return "ApiQR {}".format(np.array(self.vector))

    def __repr__(self):
        return str(self)




########################################################################
# Copyright (c) 2018
# Steffen Enders <steffen<at>enders<dot>nrw>
# Daniel Plohmann <daniel.plohmann<at>mailbox<dot>org>
# All rights reserved.
########################################################################
#
#  This file is part of apiscout
#
#  apiscout is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see
#  <http://www.gnu.org/licenses/>.
#
########################################################################

import math
from operator import itemgetter


class ApiQRContext:
    """
    Class to store all context related data that is especially needed for:
    - Exports to HTML
    - Exports to PNG
    """
    data = []
    colors_format = "RGB"
    colors_white = (255, 255, 255)
    colors_dict = {
        "execution": (41, 128, 185),
        "gui":       (241, 196, 15),
        "file":      (46, 204, 113),
        "time":      (52, 152, 219),
        "memory":    (243, 156, 18), 
        "string":    (231, 76, 60), 
        "network":   (59, 216, 214), 
        "crypto":    (142, 68, 173),
        "other":     (39, 174, 96),
        "device":    (250, 206, 32),
        "system":    (192, 57, 43), 
        "registry":  (26, 188, 156),
    }

    def __init__(self, winapi1024):
        self.data = winapi1024
        if self.dimension ** 2 != len(self.data) or int(math.log(self.dimension, 2)) != math.log(self.dimension, 2):
            raise ValueError("Vector needs to have a length for which its squareroot is a power of two")

    @property
    def dimension(self):
        return int(len(self.data) ** 0.5)

    @property
    def apis(self):
        return list(zip(map(itemgetter(0), self.data), map(itemgetter(1), self.data)))

    @property
    def colors(self):
        return list(map(lambda x: self.colors_dict[x[2]], self.data))

    @property
    def empty_vector(self):
        return [0] * len(self.data)




import struct
import os
import json
import operator
import logging
import hashlib

from .utility import get_string, get_word, get_dword, get_qword
from .PeTools import PeTools
from .OrdinalHelper import OrdinalHelper

# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOG = logging.getLogger(__name__)


class ImportDescriptor:

    def __init__(self, load_table_offset, name_table_offset, lib_name, is_64bit):
        self.load_table_offset = load_table_offset
        self.name_table_offset = name_table_offset
        self.imports = {}
        self.lib_name = lib_name
        self._is_64bit = is_64bit

    def parse(self, buf):
        self.imports = {}
        imp_offset = 0
        if self._is_64bit:
            imp = get_qword(buf, self.name_table_offset + imp_offset)
        else:
            imp = get_dword(buf, self.name_table_offset + imp_offset)
        # there is no length indicated for the array of imports, so we simply continue parsing until we hit zero.
        while imp != 0:
            thunk = self.load_table_offset + imp_offset
            self.imports[thunk] = self._parse_import(buf, self.name_table_offset + imp_offset)
            imp_offset += 8 if self._is_64bit else 4
            if self._is_64bit:
                imp = get_qword(buf, self.name_table_offset + imp_offset)
            else:
                imp = get_dword(buf, self.name_table_offset + imp_offset)

    def _parse_import(self, buf, offset):
        import_offset = get_dword(buf, offset)
        # import by ordinal
        if import_offset & 0x80000000:
            ordinal = get_word(buf, offset)
            name_by_ordinal = OrdinalHelper.resolveOrdinal(self.lib_name, ordinal)
            return  {"ordinal": ordinal, "name": name_by_ordinal, "dll_name": self.lib_name}
        return {"ordinal": get_word(buf, import_offset), "name": get_string(buf, import_offset + 2), "dll_name": self.lib_name}

    def __str__(self):
        return "{} - 0x{:x} -> 0x{:x} ({})".format(self.lib_name, self.name_table_offset, self.load_table_offset, len(self.imports))


class ImportTableLoader:

    def __init__(self, buf, sample_name=""):
        self._buffer = buf
        self._is_pe = PeTools.checkPe(buf)
        self._is_64bit = self._check_64bit()
        self._sample_name = sample_name

    def _check_64bit(self):
        pe_offset = PeTools.getPeOffset(self._buffer)
        file_characteristics_offset = pe_offset + 0x18
        if len(self._buffer) >= pe_offset + 0x18 + 2:
            file_characteristics = get_word(self._buffer, file_characteristics_offset)
            return True if file_characteristics == 0x20b else False
        return False

    def _get_64bit_bonus_offset(self):
        # in case of a PE32+, we need to adjust the data directory offset by 0x10 bytes.
        return 0x10 if self._is_64bit else 0

    def _parse_descriptor_fields(self, size, values):
        descriptor = None
        # 20b == import descriptor, 32b == delay import descriptor
        if size == 20:
            descriptor = ImportDescriptor(values[4], values[0], get_string(self._buffer, values[3]), self._is_64bit)
        elif size == 32:
            descriptor = ImportDescriptor(values[3], values[4], get_string(self._buffer, values[1]), self._is_64bit)
        return descriptor

    def _parse_descriptor(self, offset, size):
        if not size in [20, 32]:
            raise RuntimeError("Import Descriptors can only be of size 20 / 32 byte.")
        else:
            values = struct.unpack("I" * int(size / 4), self._buffer[offset:offset + size])
            return self._parse_descriptor_fields(size, values)

    def _get_table(self, directory_offset, descriptor_size):
        import_address_table = {}
        if not self._is_pe:
            return import_address_table
        pe_offset = PeTools.getPeOffset(self._buffer)
        import_dd_offset = pe_offset + directory_offset + self. _get_64bit_bonus_offset()
        it_rva = get_dword(self._buffer, import_dd_offset)
        if it_rva:
            it_offset = 0
            try:
                # the number of descriptors is not indicated, so we continue parsing until we hit zero.
                descriptor = self._parse_descriptor(it_rva + it_offset, descriptor_size)
                while descriptor.load_table_offset:
                    descriptor.parse(self._buffer)
                    import_address_table.update(descriptor.imports)
                    it_offset += descriptor_size
                    descriptor = self._parse_descriptor(it_rva + it_offset, descriptor_size)
            except IndexError:
                self._sample_name = self._sample_name if self._sample_name != "" else hashlib.sha256(self._buffer).hexdigest()
                LOG.warn("Import Table parsing was incomplete on file (%s), due to IndexError. Continuing with %d extracted imports...", self._sample_name, len(import_address_table.items()))
                return import_address_table
            except struct.error:
                self._sample_name = self._sample_name if self._sample_name != "" else hashlib.sha256(self._buffer).hexdigest()
                LOG.warn("Import Table parsing was incomplete on file (%s), due to struct.error. Continuing  with %d extracted imports...", self._sample_name, len(import_address_table.items()))
                return import_address_table
        return import_address_table

    def get_import_table(self):
        return self._get_table(0x80, 20)

    def get_delay_import_table(self):
        return self._get_table(0xE0, 32)

    def __str__(self):
        return "ImportTableLoader - buffer size: {} bytes, PE: {}, 64bit: {}".format(len(self._buffer), self._is_pe, self._is_64bit)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='apiscout',
    version='1.1.3',
    description='A library for Windows API usage recovery and similarity assessment with focus on memory dumps.',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='Daniel Plohmann',
    author_email='daniel.plohmann@mailbox.org',
    url='https://github.com/danielplohmann/apiscout',
    license="BSD 2-Clause",
    packages=find_packages(exclude=('tests', 'dbs')),
    package_data={'apiscout': ['data/winapi1024v1.txt', 'data/winapi_contexts.csv', 'data/html_frame.html']},
    data_files=[
        ('', ['LICENSE']),
    ],
    install_requires=[
        'numpy',
        'lief',
        'Pillow'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Topic :: Software Development :: Disassemblers",
    ],
)
